from dataclasses import dataclass
from typing import Optional


@dataclass
class BaseTextEncodingStrategy(object):
    name: str
    binary_string_length: int


class ASCIITextEncodingStrategy(BaseTextEncodingStrategy):
    name: str = "ascii"
    binary_string_length: int = 8

    def __init__(self):
        super().__init__(self.name, self.binary_string_length)


class UTF8TextEncodingStrategy(BaseTextEncodingStrategy):
    name: str = "utf-8"
    binary_string_length: int = 24

    def __init__(self):
        super().__init__(self.name, self.binary_string_length)


@dataclass
class TextEncodingStrategies:
    ascii = ASCIITextEncodingStrategy()
    utf8 = UTF8TextEncodingStrategy()


AVAILABLE_STRATEGY_NAMES = [s for s in dir(TextEncodingStrategies) if not s.startswith("_")]  # ["ascii", "utf8"]
DEFAULT_STRATEGY_NAME = "ascii"

def set_strategy(strategy_name: Optional[str]=None):
    """Returns a text-encoding strategy by name."""
    if strategy_name:
        strategy_name = strategy_name.replace("-", "")
    if (strategy_name is None) or (strategy_name not in AVAILABLE_STRATEGY_NAMES):
        strategy_name = DEFAULT_STRATEGY_NAME
    # The strategy objects don't have any hiphens ("utf-8" is written as "utf8")
    return getattr(TextEncodingStrategies, strategy_name)



from typing import Dict, List, Union, Optional
from genespeak.converter import Converter, set_converter


def text_to_dna(text: str, schema: str = "AGCT", binary_string_length: int = 8, strategy: str="ascii", converter: Optional[Converter] = None) -> str:
    if (converter is None) or (not isinstance(converter, Converter)):
        converter = Converter(
            schema = schema,
            binary_string_length = binary_string_length,
            strategy = strategy,
        )
    dnabase_chr: List[str] = []
    text_encoded_bin2N = converter.convert_info_to_2Nbit_binary(text, length2N=converter.strategy.binary_string_length)
    dnabase_chr, _ = converter.convert_to_dnabase(text_encoded_bin2N, strategy=converter.strategy.name)
    text_as_dnabase = ''.join(dnabase_chr)
    return text_as_dnabase


def dna_to_text(dna: str, schema: str = "AGCT", binary_string_length: int = 8, strategy: str = "ascii", converter: Optional[Converter] = None) -> str:
    converter = set_converter(
        schema = schema,
        binary_string_length = binary_string_length,
        strategy = strategy,
        converter = converter,
    )

    dnabase_as_bin = converter.encoder.chr2bin
    dnabase_as_bin2 = [dnabase_as_bin.get(x_chr) for x_chr in dna]
    length2N = converter.strategy.binary_string_length
    dnabase_as_bin2N = converter.get_bin2_to_bin2N(''.join(dnabase_as_bin2), length2N=length2N)
    # num_digits = 2 # length of any('00', '01', '10', '11')
    # binary to decimal conversion: int(x, 2)
    dnabase_as_chr = [chr(int(x, 2)) for x in dnabase_as_bin2N]
    dnabase_as_text = ''.join(dnabase_as_chr)
    return dnabase_as_text



import re

def run_length_encode(dna: str) -> str:
    """Return the run length encoding of s string

    Usage:

    ```python
    dna = "A"*3 + "G"*2 + "C" + "T"*5 + "A"*6 + "GCCT"
    dna_rle = run_length_encode(dna)
    ```

    """
    counter = []
    temp_counter = []
    # print(len(dna))
    for i, e in enumerate(dna):

        if i == 0:
            temp_counter = [e, 1]
        else:
            if e == dna[i - 1]:
                temp_counter[1] += 1
            else:
                counter.append(tuple(temp_counter.copy()))
                if i <= len(dna) - 1:
                    temp_counter = [e, 1]
            if i == len(dna) - 1:
                counter.append(tuple(temp_counter.copy()))

        ## print(i, counter)

    dna_rle = ''.join([f'{c[1]}{c[0]}' for c in counter])

    return dna_rle


def run_length_decode(dna_rle: str) -> str:
    pat = r"((\d+)(\w))"
    pat = re.compile(pat)
    return ''.join([c * int(n) for g, n, c in pat.findall(dna_rle)])



from genespeak.core import dna_to_text, text_to_dna
from genespeak.converter import DNABaseEncoder, Converter
from genespeak.text_strategies import TextEncodingStrategies

try:
    from importlib import metadata
except ImportError:  # for Python<3.8
    import importlib_metadata as metadata


__title__ = __name__
__version__ = metadata.version(__title__)

__all__ = [
    "dna_to_text",
    "text_to_dna",
    "DNABaseEncoder",
    "Converter",
    "TextEncodingStrategies",
]
